/**
 * Implementation of the Radboud Reader
 * Copyright (C) 2013  Joeri de Ruiter
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include <Keypad.h>
#include <SCLib.h>
#include <Base64.h>

//#define DEBUG

// LCD commands
#define LCD_CLEAR_DISPLAY   0x01
#define LCD_CURSOR_LEFT     0x10
#define LCD_CURSOR_BOX_ON   0x0D
#define LCD_CURSOR_BOX_OFF  0x0C

// Pins
#define LCD_txPin   1
#define SC_rstPin   12
#define SC_swPin    13
#define SC_clkPin   9
#define SC_ioPin    10
#define SC_vccPin   11
#define KP_PIN1     8
#define KP_PIN2     7
#define KP_PIN3     6
#define KP_PIN4     5
#define KP_PIN5     4
#define KP_PIN6     3
#define KP_PIN7     2

// Set up keypad
// Based on keypad example (http://www.arduino.cc/playground/Code/Keypad)
const byte ROWS = 4; //four rows
const byte COLS = 3; //three columns
char keys[ROWS][COLS] = {
  {'1','2','3'},
  {'4','5','6'},
  {'7','8','9'},
  {'*','0','#'}
};

const char KEY_CANCEL = '*';
const char KEY_OK = '#';

byte rowPins[ROWS] = {KP_PIN2, KP_PIN7, KP_PIN6, KP_PIN4};
byte colPins[COLS] = {KP_PIN3, KP_PIN1, KP_PIN5};

Keypad keypad = Keypad( makeKeymap(keys), rowPins, colPins, ROWS, COLS );

// Set up LCD
// Based on http://www.arduino.cc/playground/Learning/SparkFunSerLCD
const int LCD_delay = 10;
int LCD_position = 0;
HardwareSerial LCD = Serial1;

// Set up smartcard reader
// Based on SCTest2 example

// echocard01
//uint8_t AID[10] = {0x65,0x63,0x68,0x6f,0x63,0x61,0x72,0x64,0x30,0x31};

// radboud01
uint8_t AID[9] = {0x72,0x61,0x64,0x62,0x6f,0x75,0x64,0x30,0x31};

#define INS_DATA     0x01
#define INS_INPUT    0x02


#define RESP_SEND    0x10
#define RESP_DISPLAY 0x11
#define RESP_INPUT   0x12

// Baudrate smartcard: 2600
SmartCardReader sc(SC_ioPin, SC_rstPin, SC_vccPin, SC_swPin, SC_clkPin);

void setup(){
  Serial.begin(9600);

  // put your setup code here, to run once:
  pinMode(LCD_txPin, OUTPUT);
  Serial1.begin(9600);

/*
  // Setup LCD for 4x20
  LCD.write(0x7C);   //command flag 
  LCD.write(5);      // 4 lines
  delay(LCDdelay);  
  
  LCD.write(0x7C);   //command flag 
  LCD.write(3);      // 20 characters wide
  delay(LCDdelay);  
*/
}

// Send LCD commmand
void LCDcommand(uint8_t command) {
  LCD.write(0xFE);          
  LCD.write(command);  
  delay(LCD_delay);    
}

// Move cursor to position (0-79)
void setPosition(uint8_t new_position) {
  uint8_t pos;

  if(new_position >= 0 && new_position < 20) pos = new_position;
  else if(new_position >= 20 && new_position < 40) pos = new_position + 44;
  else if(new_position >= 40 && new_position < 60) pos = new_position - 20;  
  else if(new_position >= 60 && new_position < 80) pos = new_position + 24;
  
  LCD_position = new_position;
  
#if defined(DEBUG)
  Serial.print("LCD position:");
  Serial.println(LCD_position);
#endif

  LCD.write(0xFE);        //command flag
  LCD.write(0x80 | pos);  //move cursor.
  delay(LCD_delay);
}

// Move cursor one position left
void cursorLeft() {
  if(LCD_position == 0) setPosition(79);
  else setPosition(LCD_position - 1);  
}

// Write character and update position
void LCDwrite(char c) {
  LCD.write(c);
  LCD_position = (LCD_position + 1) % 80;
}

// Clear current character
void clearChar() {
  LCDwrite(' ');
}

// Clear line (0-3)
void clearLine(uint8_t line) {
  setPosition(line * 20);
  LCD.print("                    ");
}

void clearDisplay() {
  setPosition(0);
  LCD.print("                                                                                ");
}

// Get input from the keypad and display feedback on LCD
int getInput(uint8_t buffer[], uint8_t offset, uint8_t max_input, uint8_t start_position, boolean masked) {
  int i = 0;
  setPosition(start_position);
  LCDcommand(LCD_CURSOR_BOX_ON);
            
  while(true) {
    char key = keypad.getKey();

    if (key != NO_KEY){
      switch(key) {
        case KEY_OK:
          LCDcommand(LCD_CURSOR_BOX_OFF);        
          return i;
          break;
          
        case KEY_CANCEL:
          if(i > 0 && i < max_input) {
            cursorLeft();
            clearChar();
            cursorLeft();
            i--;
          }
          else if(i == max_input) {
            clearChar();
            cursorLeft();
            i--;             
          }
          break; 
        
        default:
          if(i < max_input) {
#if defined(DEBUG)            
            Serial.println(key);
#endif            
            
            if(!masked) LCDwrite(key);
            else LCDwrite('*');
      
            buffer[i+offset] = key;
            i++;

            // Don't let the cursor get past the last position
            if(i == max_input) {
              cursorLeft();
            }
          }
          break;
      }
    }
  }
}

// Wait for OK or Cancel key press
char waitOKCancel() {
  while(true) {
    char key = keypad.getKey();

    if (key != NO_KEY && (key == KEY_OK || key == KEY_CANCEL)) {
      return key;
    }
  }  
}

// Display message
void displayMessage(uint8_t buf[], uint8_t offset, uint8_t len) {
  LCDcommand(LCD_CLEAR_DISPLAY);
  setPosition(0);
  
  if(offset == 0) {
    LCD.write(buf, len);
  }
  else {
    for(int i = offset; i < offset + len; i++) LCD.write(buf[i]);
  }
}

// Send APDU to card and put data returned in buffer
int sendAPDU(uint8_t cla, uint8_t ins, uint8_t p1, uint8_t p2, uint8_t data_buf[], uint16_t data_size) {
  APDU_t command;

  // We just use the T=0 byte transfer
  command.cla       = cla;
  command.ins       = ins;
  command.p1        = p1;
  command.p2        = p2;
  command.data_buf  = data_buf;
  command.data_size = data_size;

#if defined(DEBUG)
    Serial.println("Sending Command ...");
    sc.dumpHEX(data_buf, data_size);  
#endif
  
  uint16_t result = sc.sendAPDU(&command);  
  
#if defined(DEBUG)
    Serial.println("Receiving ...");
    Serial.println(result, HEX);  
#endif
  
  // No data to be retrieved
  if(result == 0x9000) return 0;
   
  if ((((result >> 8) & 0xFF) == 0x61) || (((result >> 8) & 0xFF) == 0x6C)) {
    uint16_t receive_size = result & 0xFF;
#if defined(DEBUG)
      Serial.println("GET RESPONSE Command ...");
#endif
      
    // Read Command result via GET RESPONSE
    command.cla       = 0x00;
    command.ins       = 0xC0;
    command.p1        = 0x00;
    command.p2        = 0x00;
    command.data_buf  = data_buf;
    command.data_size = receive_size;
      
    result = sc.sendAPDU(&command, false);
      
#if defined(DEBUG)
      Serial.print("Received ... ");
      Serial.println(result, HEX);

      sc.dumpHEX(data_buf, receive_size);
#endif
    
    return receive_size;
  }
  
  return -result;
}

void loop(){
  // Initialize display
  LCDcommand(LCD_CLEAR_DISPLAY);
  LCDcommand(LCD_CURSOR_BOX_OFF);
  setPosition(0);
  LCD.print("Radboud Reader");  

  // Try to activate card
  uint16_t bytes_received = 0;  
  uint8_t data[256];
  char input[341];
 
  bytes_received = sc.activate(data, MAX_ATR_BYTES, CLK_2MHZ);  //CLK_4MHZ
  if(bytes_received > 0) {
    setPosition(20);
    for(int i = 0; i < bytes_received; i++)
    {
      LCD.print(data[i], HEX);
    }
  }

  int resp_size;

  // Do nothing until Serial is connected  
  while (!Serial) ;
  
  // Select application  
  resp_size = sendAPDU(0x00, 0xA4, 0x04, 0x00, AID, 9);
  
  // Get data from PC  
  Serial.setTimeout(30000);
  
  while(Serial) {
    int input_len = Serial.readBytesUntil(0x20, input, sizeof(input));  
    uint16_t len = base64_decode((char*)data, input, input_len);
    
    // Send received data to the card
    resp_size = sendAPDU(0x00, INS_DATA, 0x00, 0x00, data, len);
  
    while(resp_size > 0) {
      switch(data[0]) {
        case RESP_DISPLAY:
          // Display response by card on LCD    
          displayMessage(data, 1, resp_size - 1);
          data[0] = waitOKCancel();
          clearDisplay();
          resp_size = sendAPDU(0x00, INS_INPUT, 0x00, 0x00, data, 1);    
          break;
          
        case RESP_INPUT:
          // Get input from the user
          displayMessage(data, 3, resp_size - 3);
          resp_size = getInput(data, 1, data[2], resp_size - 3, data[1]);
          clearDisplay();
          data[0] = KEY_OK;
          resp_size = sendAPDU(0x00, INS_INPUT, 0x00, 0x00, data, resp_size + 1);
          break;
          
        case RESP_SEND:
          // Remove first byte from data
          for(int i = 0; i < resp_size - 1; i++) {
            data[i] = data[i+1];
          }      
          // Send data to PC and wait for data from PC
          len = base64_encode(input, (char*)data, resp_size - 1);
          Serial.write((uint8_t*)input, len);
          resp_size = 0;
          break;
      }
    }
  }
  
  sc.deactivate();
}
